#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "ndds/ndds_cpp.h"
#include <ndds/ndds_namespace_cpp.h>
#include "ImagePayloadSupport.h"
#include "ImagePayloadPlugin.h"
#include "ImagePayload.h"

#include <time.h>
#include <limits>

/* Delete all entities */
static int publisher_shutdown(
		DDSDomainParticipant *participant)
{
	DDS_ReturnCode_t retcode;
	int status = 0;
	if (participant != NULL) {
		retcode = participant->delete_contained_entities();
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_contained_entities error %d\n", retcode);
			status = -1;
		}
		retcode = DDSTheParticipantFactory->delete_participant(participant);
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_participant error %d\n", retcode);
			status = -1;
		}
	}
	/* RTI Connext provides finalize_instance() method on
	   domain participant factory for people who want to release memory used
	   by the participant factory singleton. Uncomment the following block of
	   code for clean destruction of the singleton. */
	/*
	   retcode = DDSDomainParticipantFactory::finalize_instance();
	   if (retcode != DDS_RETCODE_OK) {
	   printf("finalize_instance error %d\n", retcode);
	   status = -1;
	   }
	 */
	return status;
}

void setupProfiles()
{
	DDS::DomainParticipantFactory * factory = DDS::DomainParticipantFactory::get_instance();

	// Update the QoS for the factory itself
	DDS::DomainParticipantFactoryQos factoryQos;
	factory->get_qos(factoryQos);

	const int profileCount = 1;
	const char * profiles[profileCount];
	profiles[0] = "/home/fz021507/Work/ImageTopic/inheritance/Share/qos.xml";
	//profiles[1] = "/home/fz021507/Work/ImageTopic/SizeAnalysis/image_qos.xml";

	factoryQos.profile.url_profile.from_array(profiles, profileCount);
	//factoryQos.profile.ignore_resource_profile = DDS_BOOLEAN_TRUE; // Don't allow native env override
	//factoryQos.profile.ignore_user_profile = DDS_BOOLEAN_TRUE; // Don't allow cwd override
	DDS::ReturnCode_t rc = factory->set_qos(factoryQos);
	if(rc != DDS_RETCODE_OK)
	{
		printf("failed in setting qos\n");
		exit(-1);
	}
}

inline double makeTimer(const struct timespec & ts)
{
	return (double) ts.tv_sec + ((double) ts.tv_nsec * 1.0E-9);
}

inline double getTimer()
{
	struct timespec ts;
	clock_gettime(CLOCK_MONOTONIC, &ts);
	return makeTimer(ts);
}

double waitForDiscovery(Image1DataWriter *image_writer, unsigned int count, double timeout=-1.0)
{
	double start = getTimer();
	double now;
	printf("UK: Waiting for discovery of data reader(s)\n");

	DDS::PublicationMatchedStatus status;
	if (timeout < 0.0)
		timeout = std::numeric_limits<double>::max();
	int readerCount = 0;
	do
	{
		image_writer->get_publication_matched_status(status);
		readerCount = status.current_count;
		if(readerCount >= count)
		{
			printf("UK: reader count = %d\n", readerCount);
			return now-start;
		}
		::usleep(10*1000);
		now = getTimer();
	}
	while((now - start) < timeout);

	return -1.0;
}

extern "C" int publisher_main(int domainId, int sample_count, int publisher_id)
{
	DDSDomainParticipant *participant = NULL;
	DDSPublisher *publisher = NULL;
	DDSTopic *topic = NULL;
	DDSDataWriter *writer = NULL;

	Image1DataWriter *image_writer = NULL;
	Image1 *instance = NULL;

	DDS_ReturnCode_t retcode;
	DDS_InstanceHandle_t instance_handle = DDS_HANDLE_NIL;
	const char *type_name = NULL;
	int count = 0;  
	struct DDS_Duration_t send_period = {4,0}; // 4 secs, 0 nanoseconds

	setupProfiles();

	// from obelus_dds
	// Load the participant qos settings so we can manipulate them slightly
	DDS::DomainParticipantQos participantQos;
	DDSTheParticipantFactory->get_participant_qos_from_profile(
			participantQos,
			"CtQosLib", // library name
			"Image1Profile" // profile name
			);
	// disable shared memory regardless of qos configuration..
	// TODO.. need to check on why this is done
	participantQos.transport_builtin.mask = DDS::TRANSPORTBUILTIN_UDPv4;
	participant = DDSTheParticipantFactory->create_participant(
			domainId,
			participantQos,
			NULL,
			DDS::STATUS_MASK_NONE);

	if (participant == NULL) {
		printf("create_participant error\n");
		publisher_shutdown(participant);
		return -1;
	}

	/* To customize publisher QoS, use
	   participant->get_default_publisher_qos() instead */
	/*
	publisher = participant->create_publisher(
			DDS_PUBLISHER_QOS_DEFAULT, NULL, DDS_STATUS_MASK_NONE);
	*/

	publisher = participant->create_publisher_with_profile(
			"CtQosLib",
			"Image1Profile",
			NULL,
			DDS_STATUS_MASK_NONE);

	if (publisher == NULL) {
		printf("create_publisher error\n");
		publisher_shutdown(participant);
		return -1;
	}

	/* Register type before creating topic */
	type_name = Image1TypeSupport::get_type_name();
	retcode = Image1TypeSupport::register_type(
			participant, type_name);
	if (retcode != DDS_RETCODE_OK) {
		printf("register_type error %d\n", retcode);
		publisher_shutdown(participant);
		return -1;
	}

	/* To customize topic QoS, use
	   participant->get_default_topic_qos() instead */
	/*
	topic = participant->create_topic(
			"Example ReconImageData",
			type_name, DDS_TOPIC_QOS_DEFAULT, NULL / listener /,
			DDS_STATUS_MASK_NONE);
			*/

	topic = participant->create_topic_with_profile(
			"Image1Topic",
			type_name,
			"CtQosLib",
			"Image1Profile",
			NULL,
			DDS::STATUS_MASK_NONE);

	if (topic == NULL) {
		printf("create_topic error\n");
		publisher_shutdown(participant);
		return -1;
	}

	/* To customize data writer QoS, use
	   publisher->get_default_datawriter_qos() instead */
	//printf("UK: pubslisher side type name = %s\n", type_name);
	/*
	writer = publisher->create_datawriter(
			topic, DDS_DATAWRITER_QOS_DEFAULT, NULL / listener /,
			DDS_STATUS_MASK_NONE);
			*/
	writer = publisher->create_datawriter_with_profile(
			topic,
			"CtQosLib",
			"Image1Profile",
			NULL,
			DDS::STATUS_MASK_NONE);

	if (writer == NULL) {
		printf("create_datawriter error\n");
		publisher_shutdown(participant);
		return -1;
	}

	//ReconImageData_writer = Image::ReconImage::DataDataWriter::narrow(writer);
	image_writer = Image1DataWriter::narrow(writer);

	if (image_writer== NULL) {
		printf("DataWriter narrow error\n");
		publisher_shutdown(participant);
		return -1;
	}

	// print size
	int buffSize;
	PRESTypePluginEndpointData endpoint_data;
	buffSize = Image1Plugin_get_serialized_sample_max_size(
			endpoint_data,
			RTI_FALSE, // dont include encapsulation
			RTI_CDR_ENCAPSULATION_ID_CDR_NATIVE,
			0 // current alignment
			);

	printf("UK: max serialize size of Image1 = %d\n", buffSize);

	waitForDiscovery(image_writer, 1);

	/* Create data sample for writing */
	Image1 obj;
	instance = &obj;

	instance = Image1TypeSupport::create_data();
	if (instance == NULL) {
		printf("Image1TypeSupport::create_data error\n");
		publisher_shutdown(participant);
		return -1;
	}


	/* For data type that has key, if the same instance is going to be
	   written multiple times, initialize the key here
	   and register the keyed instance prior to writing */
	/*
	   instance_handle = HelloWorld_writer->register_instance(*instance);
	 */

	/* Main loop */
	//sample_count = 10;
	for (count=0; (sample_count == 0) || (count < sample_count); ++count) {
		printf("Writing Image1, count %d\n", count);
		/*
		   instance->job_type = count;

		   if ( count % 3 == 1) {
		   instance->data1._d = SHORT_MEMBER_DISCRIMINATOR;
		   instance->data1._u.short_member = (short) 1;
		   }
		   if ( count % 3 == 2) {
		   instance->data1._d = LONG_MEMBER_DISCRIMINATOR;
		   instance->data1._u.long_member = 2;
		   }



		instance->image_key.exam = 1;
		instance->image_key.exam_unique = 1;
		instance->image_key.series = 1;
		instance->image_key.series_unique = 1;
		instance->image_key.image_unique = 1;
		instance->image_key.image_number = 1;
		std::string img_path = "somePath";
		std::string suite_id = "some_id";
		instance->image_key.image_path = const_cast<char*>(img_path.c_str());
		instance->image_key.suite_id = const_cast<char*>(suite_id.c_str());
		instance->image_key.image_allocation_time = 1;
		std::string img_uid = "some_uid";
		instance->image_key.image_uid.from_array(const_cast<char*>(img_uid.c_str()), img_uid.length());



		char *scan_key = new char[256];
		char *recon_req = new char[3072];
		char *rpeak = new char[11264];


		instance->scan_key.maximum(0);
		bool res = instance->scan_key.loan_contiguous((DDS_Octet*)scan_key, 256, 256);
		if(!res)
		{
			printf("Failed to loan memory to dds message\n");
		}


		instance->recon_request.maximum(0);
		res = instance->recon_request.loan_contiguous((DDS_Octet*)recon_req, 3072, 3072);
		if(!res)
		{
			printf("Failed to loan memory to dds message\n");
		}


		instance->rpeak_info.maximum(0);
		res = instance->rpeak_info.loan_contiguous((DDS_Octet*)rpeak, 11264, 11264);
		if(!res)
		{
			printf("Failed to loan memory to dds message\n");
		}


		short *raw_image = new short[512*512];
		instance->image_data.data.maximum(0);
		res = instance->image_data.data.loan_contiguous((DDS_Short*)raw_image, 512*512, 512*512);


		instance->additional_data1._u.dummy = false;
		instance->additional_data1._d = Image::RECON_IMAGE_ADDTYPE_NONE;

		instance->additional_data2._u.dummy = false;
		instance->additional_data2._d = Image::RECON_IMAGE_ADDTYPE_NONE;

		instance->additional_data3._u.dummy = false;
		instance->additional_data3._d = Image::RECON_IMAGE_ADDTYPE_NONE;










		//Image::ReconImage::DataTypeSupport::print_data(instance);
		//printf("UK: size of scan_key = %d %d\n", instance->scan_key.length(), scan_key.length());


		int buffSize;
		PRESTypePluginEndpointData endpoint_data;
		buffSize = Image::ReconImage::DataPlugin_get_serialized_sample_size(
				endpoint_data,
				RTI_FALSE,
				RTI_CDR_ENCAPSULATION_ID_CDR_NATIVE,
				0,
				instance);
		printf("serialized buff size = %d\n", buffSize);



		retcode = ReconImageData_writer->write(*instance, instance_handle);
		if (retcode != DDS_RETCODE_OK) {
			printf("write error %d\n", retcode);
		}
		NDDSUtility::sleep(send_period);


		instance->scan_key.unloan();
		instance->recon_request.unloan();
		instance->rpeak_info.unloan();
		instance->image_data.data.unloan();
		 */

		retcode = image_writer->write(*instance, instance_handle);
	}

	/*
	   retcode = HelloWorld_writer->unregister_instance(
	 *instance, instance_handle);
	 if (retcode != DDS_RETCODE_OK) {
	 printf("unregister instance error %d\n", retcode);
	 }
	 */
	/* Delete data sample */
	//retcode = Image::ReconImage::DataTypeSupport::delete_data(instance);
	//retcode = Image1TypeSupport::delete_data(instance);
	//if (retcode != DDS_RETCODE_OK) {
	//	printf("Image1TypeSupport::delete_data error %d\n", retcode);
	//}
	/* Delete all entities */
	return publisher_shutdown(participant);
}

int main(int argc, char *argv[])
{
	int domainId = 0;
	int sample_count = 0; /* infinite loop */
	int publisher_id;

	/*
	   if (argc >= 2)
	   {
	   publisher_id = atoi(argv[1]);
	   }
	   if (argc >= 2) {
	   domainId = atoi(argv[1]);
	   }
	   if (argc >= 3) {
	   sample_count = atoi(argv[2]);
	   }
	 */

	/* Comment this to turn off additional logging */
	NDDSConfigLogger::get_instance()->
		set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API, 
				NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);

	NDDSConfigLogger::get_instance()->
		set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_COMMUNICATION,
				NDDS_CONFIG_LOG_VERBOSITY_WARNING);

	return publisher_main(domainId, sample_count, publisher_id);
}
