#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "ndds/ndds_cpp.h"
#include <ndds/ndds_namespace_cpp.h>
#include "ImagePayloadSupport.h"
#include "ImagePayloadPlugin.h"
#include "ImagePayload.h"
#include <sstream>

class ConsolidatedImageListener : public DDSDataReaderListener {
	public:
		virtual void on_requested_deadline_missed(
				DDSDataReader* /*reader*/,
				const DDS_RequestedDeadlineMissedStatus& /*status*/) {}

		virtual void on_requested_incompatible_qos(
				DDSDataReader* /*reader*/,
				const DDS_RequestedIncompatibleQosStatus& /*status*/) {}

		virtual void on_sample_rejected(
				DDSDataReader* /*reader*/,
				const DDS_SampleRejectedStatus& /*status*/) {}
		virtual void on_liveliness_changed(
				DDSDataReader* /*reader*/,
				const DDS_LivelinessChangedStatus& /*status*/) {}
		virtual void on_sample_lost(
				DDSDataReader* /*reader*/,
				const DDS_SampleLostStatus& /*status*/) {}
		virtual void on_subscription_matched(
				DDSDataReader* /*reader*/,
				const DDS_SubscriptionMatchedStatus& /*status*/) {}
		virtual void on_data_available(DDSDataReader* reader);
};

void ConsolidatedImageListener::on_data_available(DDSDataReader* reader)
{
	ConsolidatedImageDataReader *image_reader = NULL;
	ConsolidatedImageSeq data_seq;
	DDS_SampleInfoSeq info_seq;
	DDS_ReturnCode_t retcode;
	int i;

	image_reader = ConsolidatedImageDataReader::narrow(reader);
	if (image_reader == NULL) {
		printf("DataReader narrow error\n");
		return;
	}
	retcode = image_reader->take(
			data_seq, 
			info_seq, 
			DDS_LENGTH_UNLIMITED,
			DDS_ANY_SAMPLE_STATE, 
			DDS_ANY_VIEW_STATE, 
			DDS_ANY_INSTANCE_STATE);

	if (retcode == DDS_RETCODE_NO_DATA) {
		printf("no data\n");
		return;
	} else if (retcode != DDS_RETCODE_OK) {
		printf("take error %d\n", retcode);
		return;
	}
	printf("UK: length = %d\n", data_seq.length());
	for (i = 0; i < data_seq.length(); ++i) {
		if (info_seq[i].valid_data) {
			//ConsolidatedImageTypeSupport::print_data(&data_seq[i]);

			{
				int buffSize;
				PRESTypePluginEndpointData endpoint_data;
				buffSize = ConsolidatedImagePlugin_get_serialized_sample_size(
						endpoint_data,
						RTI_FALSE,
						RTI_CDR_ENCAPSULATION_ID_CDR_NATIVE,
						0,
						&(data_seq[i]));
				printf("serialized buff size (subscriber) = %d\n", buffSize);
			}
		}
	}
	retcode = image_reader->return_loan(data_seq, info_seq);
	if (retcode != DDS_RETCODE_OK) {
		printf("return loan error %d\n", retcode);
	}
}

/* Delete all entities */
static int subscriber_shutdown(
		DDSDomainParticipant *participant)
{
	DDS_ReturnCode_t retcode;
	int status = 0;
	if (participant != NULL) {
		retcode = participant->delete_contained_entities();
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_contained_entities error %d\n", retcode);
			status = -1;
		}
		retcode = DDSTheParticipantFactory->delete_participant(participant);
		if (retcode != DDS_RETCODE_OK) {
			printf("delete_participant error %d\n", retcode);
			status = -1;
		}
	}
	/* RTI Connext provides the finalize_instance() method on
	   domain participant factory for people who want to release memory used
	   by the participant factory. Uncomment the following block of code for
	   clean destruction of the singleton. */
	/*
	   retcode = DDSDomainParticipantFactory::finalize_instance();
	   if (retcode != DDS_RETCODE_OK) {
	   printf("finalize_instance error %d\n", retcode);
	   status = -1;
	   }
	 */
	return status;
}

void setupProfiles()
{
	DDS::DomainParticipantFactory * factory = DDS::DomainParticipantFactory::get_instance();

	// Update the QoS for the factory itself
	DDS::DomainParticipantFactoryQos factoryQos;
	factory->get_qos(factoryQos);

	const int profileCount = 1;
	const char * profiles[profileCount];
	profiles[0] = "/home/fz021507/Work/ImageTopic/inheritance/Share/qos.xml";
	//profiles[1] = "/home/fz021507/Work/ImageTopic/SizeAnalysis/image_qos.xml";

	factoryQos.profile.url_profile.from_array(profiles, profileCount);
	//factoryQos.profile.ignore_resource_profile = DDS_BOOLEAN_TRUE; // Don't allow native env override
	//factoryQos.profile.ignore_user_profile = DDS_BOOLEAN_TRUE; // Don't allow cwd override
	DDS::ReturnCode_t rc = factory->set_qos(factoryQos);
	if(rc != DDS_RETCODE_OK)
	{
		printf("failed in setting qos\n");
		exit(-1);
	}
}

extern "C" int subscriber_main(int domainId, int sample_count)
{
	DDSDomainParticipant *participant = NULL;
	DDSSubscriber *subscriber = NULL;
	DDSTopic *topic = NULL;
	ConsolidatedImageListener *reader_listener = NULL; 
	DDSDataReader *reader = NULL;
	DDS_ReturnCode_t retcode;
	const char *type_name = NULL;
	int count = 0;
	DDS_Duration_t receive_period = {4,0};
	int status = 0;

	setupProfiles();

	// from obelus_dds
	// Load the participant qos settings so we can manipulate them slightly
	DDS::DomainParticipantQos participantQos;
	DDSTheParticipantFactory->get_participant_qos_from_profile(
			participantQos,
			"CtQosLib", // library name
			"ConsolidatedImageProfile" // profile name
			);
	// disable shared memory regardless of qos configuration..
	// TODO.. need to check on why this is done
	participantQos.transport_builtin.mask = DDS::TRANSPORTBUILTIN_UDPv4;
	participant = DDSTheParticipantFactory->create_participant(
			domainId,
			participantQos,
			NULL,
			DDS::STATUS_MASK_NONE);

	if (participant == NULL) {
		printf("create_participant error\n");
		subscriber_shutdown(participant);
		return -1;
	}

	/* To customize the subscriber QoS, use
	   participant->get_default_subscriber_qos() */
	subscriber = participant->create_subscriber_with_profile(
			"CtQosLib",
			"ConsolidatedImageProfile",
			NULL,
			DDS_STATUS_MASK_NONE);

	if (subscriber == NULL) {
		printf("create_subscriber error\n");
		subscriber_shutdown(participant);
		return -1;
	}

	/* Register the type before creating the topic */
	type_name = ConsolidatedImageTypeSupport::get_type_name();
	retcode = ConsolidatedImageTypeSupport::register_type(
			participant, type_name);
	if (retcode != DDS_RETCODE_OK) {
		printf("register_type error %d\n", retcode);
		subscriber_shutdown(participant);
		return -1;
	}

	topic = participant->create_topic_with_profile(
			"ConsolidatedImageTopic",
			type_name,
			"CtQosLib",
			"ConsolidatedImageProfile",
			NULL,
			DDS::STATUS_MASK_NONE);

	if (topic == NULL) {
		printf("create_topic error\n");
		subscriber_shutdown(participant);
		return -1;
	}

	/* Create a data reader listener */
	reader_listener = new ConsolidatedImageListener();

	reader = subscriber->create_datareader_with_profile(
			topic,
			"CtQosLib",
			"ConsolidatedImageProfile",
			reader_listener,
			DDS_STATUS_MASK_ALL);

	if (reader == NULL) {
		printf("create_datareader error\n");
		subscriber_shutdown(participant);
		delete reader_listener;
		return -1;
	}

	/* Main loop */
	for (count=0; (sample_count == 0) || (count < sample_count); ++count) {
		printf("ConsolidatedImage subscriber sleeping for %d sec...\n",
				receive_period.sec);
		NDDSUtility::sleep(receive_period);
	}
	/* Delete all entities */
	status = subscriber_shutdown(participant);
	delete reader_listener;
	return status;
}

int main(int argc, char *argv[])
{
	int domainId = 0;
	int sample_count = 0; /* infinite loop */
	if (argc >= 2) {
		domainId = atoi(argv[1]);
	}
	if (argc >= 3) {
		sample_count = atoi(argv[2]);
	}
	/* Uncomment this to turn on additional logging */
	NDDSConfigLogger::get_instance()->
		set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API, 
				NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);

	NDDSConfigLogger::get_instance()->
		set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_COMMUNICATION,
				NDDS_CONFIG_LOG_VERBOSITY_WARNING);

	return subscriber_main(domainId, sample_count);
}
